# RiboViz and CWL

## Introduction

This page describes how to run a prototype of a subset of the RiboViz workflow implemented in [Common Workflow Language](http://www.commonwl.org) (CWL) and a discussion on the pros and cons of using CWL to implement the RiboViz workflow.

---

## Common Workflow Language links

* [Common Workflow Language](http://www.commonwl.org)
* [GitHub](https://github.com/common-workflow-language)
* [cwltool](https://github.com/common-workflow-language/cwltool)

---

## Folder contents

* [cutadapt.cwl](./cutadapt.cwl): tool wrapper for `cutadapt`.
* [cutadapt-job.yml](./cutadapt-job.yml): job configuration for above.
* [riboviz-workflow.cwl](./riboviz-workflow.cwl): workflow invoking `hisat2-build` and `cutadapt`.
* [riboviz-job.yml](./riboviz-job.yml): job configuration for the above.
* [CommonWorkflowLanguage.md](./CommonWorkflowLanguage.md): Notes on CWL.
* [SnakemakeCwl.md](./SnakemakeCwl.md): Notes on Snakemake's `--export-cwl` option.

---

## Run CWL jobs via cwltool

Create a new conda environment from the current RiboViz one and activate it:

```console
$ conda create --name riboviz-cwltool --clone riboviz
$ conda activate riboviz-cwltool
```

Install CWL reference implementation and node.js:

```console
$ sudo apt-get -y install nodejs
$ pip install cwlref-runner
$ cwl-runner --version
/home/ubuntu/miniconda3/envs/riboviz-cwltool/bin/cwl-runner 2.0.20200224214940
$ cwltool --version
/home/ubuntu/miniconda3/envs/riboviz-cwltool/bin/cwltool 2.0.20200224214940
```

Install html5lib-python:

* [Website](https://html5lib.readthedocs.io/en/latest/)
* [GitHub](https://github.com/html5lib/html5lib-python)

```console
$ conda install html5lib
```

This is optional but will avoid warnings of form:

```
Could not load extension schema
https://schema.org/docs/schema_org_rdfa.html: html5lib is not
installed, cannot use RDFa and Microdata parsers.
```

Copy over `cutadapt` tool wrapper, workflow with `cutadapt` and `hisat2-build` steps and related job configuration files:

```console
$ cd riboviz
$ cp ~/workflows/cwl/*cwl .
$ cp ~/workflows/cwl/*yml .
```

Get third-party `hisat2-build` tool wrapper:

```console
$ wget https://raw.githubusercontent.com/pitagora-network/pitagora-cwl/master/tools/hisat2/index/hisat2_index.cwl
```

Run `cutadapt` example:

```console
$ cwltool --no-container cutadapt.cwl  cutadapt-job.yml
```

Alternatively:

```console
$ cwltool --no-container cutadapt.cwl --adapter CTGTAGGCACC --trim --overlap 1 --min_length 1 --cores 0 --input vignette/input/SRR1042855_s1mi.fastq.gz --output trim.fq
```

Run sample workflow:

```console
$ cwl-runner --no-container riboviz-workflow.cwl riboviz-job.yml
```

---

## Assessment

### Ease of download, install, tutorials, initial use

The [Common Workflow Language User Guide](https://www.commonwl.org/user_guide/) provides a readable, step-by-step, [Software Carpentry](https://software-carpentry.org)-style, introduction to the language.

I run the "hello world!" example ([First Example](https://www.commonwl.org/user_guide/02-1st-example/index.html)) using both the Common Workflow Language reference implementation, [cwltool](https://github.com/common-workflow-language/cwltool), and the implementation in the [Toil](https://toil.readthedocs.io) Python workflow engine. To download and run these took less than 30 minutes.

### Ease of implementation of key RiboViz steps

A lot of preparatory reading is required to go beyond the "hello world!" example. The User Guide took about 4 hours to work through, running all the examples.

It took half a day to implement a very basic workflow with indexing (using a third-party `hisat2-build` tool wrapper) and trimming of a single sample (using a `cutadapt` tool wrapper written myself).

Constructing file names from sample tags requires embedding JavaScript into YAML (or JSON) workflow documents. This would incur an additional language for RiboViz developers to be familiar with (in addition to Python and R).

The "edit-compile-run" development cycle was very slow.

Tool wrappers can be verbose for what are single command-line invocations e.g. the basic `cutadapt` wrapper in `cutadapt.cwl` is 47 lines.

Errors in CWL syntax or during execution can lead to cryptic error messages or require a lot of investigation to track down their causes.

### Sample-specific sub-directories

Untried as even getting a basic workflow running, with all files being output to a common directory took time.

### Parse YAML configuration files

All CWL documents, including job files with configuration for a workflow, can be written in YAML or JSON.

### Dry run option, validating configuration and input file existence

Untried.

### Tool/step-specific log files

Untried.

### Output bash script, that can be rerun

Untried.

### Access configuration file name from within CWL

Untried.

### If processing of one sample fails will the rest be processed?

Untried.

### Conditional behaviour

For example, for selecting to create index file, bedgraphs, UMI group summary files, to deduplicate or count reads.

CWL issues on [Proposal for conditionals #494](https://github.com/common-workflow-language/common-workflow-language/issues/494), [Proposal: conditionals II (refinement of run_if) #854](https://github.com/common-workflow-language/common-workflow-language/issues/854), [Collecting use cases for workflow level conditionals #725](https://github.com/common-workflow-language/common-workflow-language/issues/725) exist and were added, in February 2020, to a [1.2](https://github.com/common-workflow-language/common-workflow-language/milestone/9) milestone. This has no due date.

### Other

Any additional requirements or functionality that we might require would most likely incur changes to both the specification and any implementations and so require liaison with the CWL stakeholders. Given the number of stakeholders change would not necessarily be rapid.

A colleage at EPCC evaluated CWL about a year and a half ago. They commented that "Simple workflows showed promise, but more complicated examples ... were not possible as the current CWL specification does not provide conditional workflow paths (such as being able to check output states and re-run/loop back over previous steps)."
